use std::{cell::RefCell, rc::Rc};

use super::{
    block::CommandBlock,
    event::Event,
    token::Token,
    utils::{describe_line, parse_line},
    Pair,
};

#[derive(Debug, Clone)]
#[allow(dead_code)]
pub enum Command {
    Say { name: Option<String>, text: String },
    Eat { food: String, politely: bool },
    Define { variable: String, value: Token },
}

impl Command {
    pub fn execute(&self, context: &Rc<RefCell<CommandBlock>>) -> Option<Event> {
        use Command::*;
        Some(
            match self {
                Say { name, text } => Event::Say {
                    name: name.clone(),
                    text: text.clone(),
                },
                Define { .. } => panic!("define command should not be executed at runtime!"),
                Eat { .. } => return None,
            }
            .process(context),
        )
    }
}

pub fn parse_command(pair: Pair) -> Command {
    use Command::*;
    use Token::*;
    let line = parse_line(pair);
    macro_rules! unknown {
        () => {
            panic!("Unknown command {}", describe_line(&line))
        };
    }
    match line.as_slice() {
        [Str(text)] => Say {
            name: None,
            text: text.to_owned(),
        },
        [Str(name), Str(text)] => Say {
            name: Some(name.to_owned()),
            text: text.to_owned(),
        },
        [Keyword(keyword), Keyword(variable), Keyword(equals), value]
            if keyword.eq("define") && equals.eq("=") =>
        {
            Define {
                variable: variable.to_owned(),
                value: value.clone(),
            }
        }
        [Keyword(keyword), Str(food), tail @ ..] if keyword.eq("eat") => Eat {
            food: food.to_owned(),
            politely: match tail {
                [Boolean(politely)] => *politely,
                _ => unknown!(),
            },
        },
        _ => unknown!(),
    }
}
